"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var CompressedJSON_1 = require("./CompressedJSON");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var MIN_LENGTH_FOR_ENUM = 10;
function concatArrays(arrays) {
    var combined = [];
    for (var i = 0; i < arrays.length; i++)
        combined = combined.concat(arrays[i]);
    return combined;
}
var TypeInference = /** @class */ (function () {
    function TypeInference(_inferMaps, _inferEnums) {
        var _this = this;
        this._inferMaps = _inferMaps;
        this._inferEnums = _inferEnums;
        this._typeBuilder = new TypeBuilder_1.TypeBuilder();
        this.inferType = function (cjson, typeName, valueArray) {
            var haveNull = false;
            var haveBool = false;
            var haveInteger = false;
            var haveDouble = false;
            var haveString = false;
            var enumCaseMap = {};
            var enumCases = [];
            var objects = [];
            var arrays = [];
            for (var _i = 0, valueArray_1 = valueArray; _i < valueArray_1.length; _i++) {
                var value = valueArray_1[_i];
                var t = CompressedJSON_1.valueTag(value);
                switch (t) {
                    case CompressedJSON_1.Tag.Null:
                        haveNull = true;
                        break;
                    case CompressedJSON_1.Tag.False:
                    case CompressedJSON_1.Tag.True:
                        haveBool = true;
                        break;
                    case CompressedJSON_1.Tag.Integer:
                        haveInteger = true;
                        break;
                    case CompressedJSON_1.Tag.Double:
                        haveDouble = true;
                        break;
                    case CompressedJSON_1.Tag.InternedString:
                        if (!haveString && valueArray.length >= MIN_LENGTH_FOR_ENUM) {
                            var s = cjson.getStringForValue(value);
                            if (!Object.prototype.hasOwnProperty.call(enumCaseMap, s)) {
                                enumCaseMap[s] = 0;
                                enumCases.push(s);
                            }
                            enumCaseMap[s] += 1;
                        }
                        else {
                            haveString = true;
                        }
                        break;
                    case CompressedJSON_1.Tag.UninternedString:
                        if (!haveString) {
                            haveString = true;
                            enumCaseMap = {};
                            enumCases = [];
                        }
                        break;
                    case CompressedJSON_1.Tag.Object:
                        objects.push(cjson.getObjectForValue(value));
                        break;
                    case CompressedJSON_1.Tag.Array:
                        arrays.push(cjson.getArrayForValue(value));
                        break;
                    default:
                        return Support_1.assertNever(t);
                }
            }
            var types = [];
            if (haveNull) {
                types.push(_this._typeBuilder.getPrimitiveType("null"));
            }
            if (haveBool) {
                types.push(_this._typeBuilder.getPrimitiveType("bool"));
            }
            if (haveDouble) {
                types.push(_this._typeBuilder.getPrimitiveType("double"));
            }
            else if (haveInteger) {
                types.push(_this._typeBuilder.getPrimitiveType("integer"));
            }
            if (_this._inferEnums && enumCases.length > 0 && enumCases.length < Math.sqrt(valueArray.length)) {
                types.push(_this._typeBuilder.getEnumType(typeName, immutable_1.OrderedSet(enumCases)));
            }
            else if (enumCases.length > 0 || haveString) {
                types.push(_this._typeBuilder.getPrimitiveType("string"));
            }
            if (objects.length > 0) {
                types.push(_this.inferClassType(cjson, typeName, objects));
            }
            if (arrays.length > 0) {
                var combined = concatArrays(arrays);
                types.push(_this._typeBuilder.getArrayType(_this.inferType(cjson, pluralize.singular(typeName), combined)));
            }
            if (types.length === 0) {
                return _this._typeBuilder.getPrimitiveType("any");
            }
            if (types.length === 1) {
                return types[0];
            }
            return _this._typeBuilder.getUnionType(typeName, immutable_1.OrderedSet(types));
        };
        this.inferClassType = function (cjson, typeName, objects) {
            var combined = concatArrays(objects);
            var propertyNames = [];
            var propertyValues = {};
            for (var i = 0; i < combined.length; i += 2) {
                var key = cjson.getStringForValue(combined[i]);
                var value = combined[i + 1];
                if (!Object.prototype.hasOwnProperty.call(propertyValues, key)) {
                    propertyNames.push(key);
                    propertyValues[key] = [];
                }
                propertyValues[key].push(value);
            }
            var properties = [];
            var couldBeMap = _this._inferMaps;
            for (var _i = 0, propertyNames_1 = propertyNames; _i < propertyNames_1.length; _i++) {
                var key = propertyNames_1[_i];
                var values = propertyValues[key];
                var t = _this.inferType(cjson, key, values);
                if (values.length < objects.length) {
                    t = Type_1.makeNullable(t, key);
                }
                if (couldBeMap && properties.length > 0 && t !== properties[0][1]) {
                    couldBeMap = false;
                }
                properties.push([key, t]);
            }
            if (couldBeMap && properties.length >= 20) {
                return _this._typeBuilder.getMapType(properties[0][1]);
            }
            return _this._typeBuilder.getClassType(typeName, immutable_1.OrderedMap(properties));
        };
    }
    return TypeInference;
}());
exports.TypeInference = TypeInference;
